"""
A multiprocessing context where ``proc.daemon`` is always False,
and setting it to True is ignored.
All processes will still be cleaned up though.
For cleaning up, we use SIGINT (instead of the standard SIGTERM)
such that the processes can do proper cleanup themselves.

See :class:`NonDaemonicSpawnProcess` and :class:`NonDaemonicSpawnContext`.

This is when you don't want to use the default multiprocessing fork start method,
but the spawn start method, but you also want that the started processes are never daemonic.

References:
    https://github.com/rwth-i6/returnn/issues/1494
    https://github.com/rwth-i6/returnn/issues/1495
    https://github.com/pytorch/pytorch/issues/15950
"""

from __future__ import annotations
from typing import Optional, Any, Callable, Tuple
import os
import signal
import atexit

# noinspection PyProtectedMember
from multiprocessing.context import BaseContext, SpawnProcess

# noinspection PyUnresolvedReferences
from multiprocessing.util import is_exiting


class NonDaemonicSpawnProcess(SpawnProcess):
    """
    This process is always non-daemon, even if ``proc.daemon=True`` is executed
    (like https://stackoverflow.com/a/8963618/133374).

    Still, we make sure that the proc is cleaned up at exit.
    Instead of using SIGTERM as Python does for normal daemonic threads,
    we use SIGINT, to allow the subprocess to do proper cleanup,
    e.g. like cleaning up sub-sub procs.
    So the sub proc would not leave its children orphaned.
    Note, if SIGINT does nothing on the subproc, this will hang.
    """

    daemon = property(lambda self: False, lambda self, v: None)  # always False

    pre_init_func: Optional[Callable[[], None]] = None

    _at_exit_cleanup_handler: Optional[_AtExitCleanupProcess] = None

    def start(self):
        """start"""
        super().start()
        self._at_exit_cleanup_handler = _AtExitCleanupProcess(self.ident)
        atexit.register(self._at_exit_cleanup_handler)

    def terminate(self):
        """terminate"""
        super().terminate()
        self._close_cleanup()

    def kill(self):
        """kill"""
        super().kill()
        self._close_cleanup()

    def join(self, timeout=None):
        """join"""
        if is_exiting():
            # Process._bootstrap (the subproc main logic) will do an early call to
            # multiprocessing.util._exit_function, which terminates all daemon procs,
            # and then joins all procs (daemon or non-daemon).
            # As we are all non-daemonic here, it means it will call join()
            # without having send SIGINT, SIGTERM, SIGKILL or anything to the proc
            # -- so this will just hang.
            # is_exiting() will be True exactly if this _exit_function was called.
            # If the proc is still alive, send SIGINT now,
            # just like our atexit handler would do.
            # However, our atexit handler will only run at some later point,
            # but then it is too late, as we would hang here now in the join().
            # https://github.com/rwth-i6/returnn/issues/1497
            # https://github.com/python/cpython/issues/114220
            try:
                os.kill(self.ident, signal.SIGINT)
            except ProcessLookupError:
                pass
        super().join(timeout=timeout)
        self._close_cleanup()

    def close(self):
        """close"""
        super().close()
        self._close_cleanup()

    def _close_cleanup(self):
        """close"""
        if not self._at_exit_cleanup_handler:
            return
        atexit.unregister(self._at_exit_cleanup_handler)
        self._at_exit_cleanup_handler = None

    def __reduce__(self):
        res_tuple = super().__reduce__()
        if not self.pre_init_func:
            return res_tuple
        reconstruct_func, reconstruct_args, *other = res_tuple
        # Use our own reconstruct function to call the pre_init_func.
        # This is unpickled and executed *before* the other state is unpickled.
        # This is important: This allows to potentially prepare some global state,
        # to make the following unpickling work.
        # E.g. in case the remaining state depends on some dynamic module,
        # which must be imported in a special way before (e.g. __returnn_config__),
        # this is the way to do it.
        # Note that internally, multiprocessing SpawnProcess does sth similar,
        # see multiprocessing.spawn._main, spawn.prepare.
        return (self._reconstruct_with_pre_init_func, (reconstruct_func, reconstruct_args, self.pre_init_func)) + tuple(
            other
        )

    @staticmethod
    def _reconstruct_with_pre_init_func(
        reconstruct_func: Callable, reconstruct_args: Tuple[Any, ...], pre_init_func: Callable[[], None]
    ):
        pre_init_func()
        return reconstruct_func(*reconstruct_args)


class NonDaemonicSpawnContext(BaseContext):
    """
    Spawn start methods, where all procs are non-daemonic.
    """

    _name = "spawn_non_daemonic"

    def __init__(self, *, process_pre_init_func: Optional[Callable[[], None]] = None):
        super().__init__()
        self.process_pre_init_func = process_pre_init_func

    # noinspection PyPep8Naming
    def Process(self, *args, **kwargs):
        """create a new process"""
        proc = NonDaemonicSpawnProcess(*args, **kwargs)
        if self.process_pre_init_func:
            proc.pre_init_func = self.process_pre_init_func
        return proc


class _AtExitCleanupProcess:
    def __init__(self, proc_pid: int):
        self.cur_pid = os.getpid()
        self.proc_pid = proc_pid

    def __call__(self):
        if os.getpid() != self.cur_pid:  # e.g. in fork
            return  # ignore
        if self.proc_pid is None:  # already cleaned
            return
        # The proc might have been killed by some other code. That's ok.
        try:
            # Send SIGINT, not SIGTERM or SIGKILL. See NonDaemonicSpawnProcess docstring.
            os.kill(self.proc_pid, signal.SIGINT)
        except ProcessLookupError:
            pass
        else:
            try:
                os.waitpid(self.proc_pid, 0)
            except ChildProcessError:
                pass
        self.proc_pid = None



from __future__ import annotations

import _setup_test_env  # noqa

import logging
import sys
import os
import os.path
import tensorflow as tf
import numpy as np
import unittest
import contextlib

from returnn.tf.network import *
from returnn.tf.layers.signal_processing import *
from returnn.config import Config

from returnn.util import better_exchook

better_exchook.replace_traceback_format_tb()


@contextlib.contextmanager
def make_scope():
    """
    :rtype: tf.compat.v1.Session
    """
    with tf.Graph().as_default() as graph:
        with tf_compat.v1.Session(graph=graph) as session:
            yield session


def test_melFilterbankLayer():
    with make_scope() as session:
        n_in, n_out = 257, 3
        layer_name = "mel_filterbank_layer"
        config = Config()
        config.update(
            {
                "num_outputs": n_out,
                "num_inputs": n_in,
                "network": {
                    layer_name: {
                        "class": "mel_filterbank",
                        "fft_size": 512,
                        "nr_of_filters": n_out,
                        "n_out": n_out,
                        "is_output_layer": True,
                        "from": "data:data",
                    }
                },
            }
        )
        network = TFNetwork(config=config, train_flag=True)
        network.construct_from_dict(config.typed_value("network"))
        layer = network.layers[layer_name]
        test_out = session.run(
            layer.output.placeholder, feed_dict={network.get_extern_data("data").placeholder: np.ones((1, 1, 257))}
        )
        assert np.sum(test_out - np.asarray([28.27923584, 53.10634232, 99.71585846], dtype=np.float32)) < 1e-5


def test_complexLinearProjectionLayer():
    with make_scope() as session:
        n_in, n_out = 514, 128
        layer_name = "clp_layer"
        config = Config()
        config.update(
            {
                "num_outputs": n_out,
                "num_inputs": n_in,
                "network": {
                    layer_name: {
                        "class": "complex_linear_projection",
                        "nr_of_filters": n_out,
                        "n_out": n_out,
                        "is_output_layer": True,
                        "from": "data:data",
                    }
                },
            }
        )
        network = TFNetwork(config=config, train_flag=True)
        network.construct_from_dict(config.typed_value("network"))
        layer = network.layers[layer_name]
        assert isinstance(layer, ComplexLinearProjectionLayer)
        i_r = np.ones((1, n_in // 2))
        i_i = np.ones((1, n_in // 2)) * 0.5
        test_input = np.expand_dims(
            np.reshape(np.transpose(np.reshape(np.concatenate([i_r, i_i], axis=1), (1, 2, 257)), [0, 2, 1]), (1, 514)),
            0,
        )
        test_clp_kernel = np.ones((2, n_in // 2, 128))
        network.initialize_params(session)
        test_clp_output = session.run(
            layer.output.placeholder,
            feed_dict={network.get_extern_data("data").placeholder: test_input, layer._clp_kernel: test_clp_kernel},
        )
        assert test_clp_output[0, 0, 0] - 6.00722122 < 1e-5


def test_MultichannelStftLayer():
    def _get_ref_output(time_sig, fft_size, frame_size, frame_shift, window_name, frame_nr, channel_nr):
        import numpy as np

        frame_start = frame_nr * frame_shift
        frame_end = frame_start + frame_size
        frame = time_sig[0, frame_start:frame_end, channel_nr]
        if window_name == "hanning":
            window = np.hanning(frame_size)
        windowed_frame = window * frame
        out = np.fft.rfft(windowed_frame, fft_size)
        return out

    def test_rfftStftConfig_01():
        with make_scope() as session:
            layer_name = "stft_layer"
            fft_size = 400
            frame_size = 400
            frame_shift = 160
            window = "hanning"
            test_input = np.ones((1, 32000, 2), dtype=np.float32)
            config = Config()
            config.update(
                {
                    "num_outputs": int(fft_size / 2) + 1 * test_input.shape[2],
                    "num_inputs": test_input.shape[2],
                    "network": {
                        layer_name: {
                            "class": "multichannel_stft_layer",
                            "frame_shift": frame_shift,
                            "frame_size": frame_size,
                            "window": window,
                            "fft_size": fft_size,
                            "use_rfft": True,
                            "nr_of_channels": 2,
                            "is_output_layer": True,
                            "from": "data:data",
                        }
                    },
                }
            )
            network = TFNetwork(config=config, train_flag=True)
            network.construct_from_dict(config.typed_value("network"))
            layer = network.layers[layer_name]
            test_output = session.run(
                layer.output.placeholder, {network.get_extern_data("data").placeholder: test_input}
            )
            ref0 = _get_ref_output(test_input, fft_size, frame_size, frame_shift, window, 0, 0)
            # np.fft.rfft and tensorflow.python.ops.rfft differ a little bit in their
            # results, thus an error margin is allowed in the result
            resultDiff = np.abs(test_output[0, 0, 0 : (int(fft_size / 2) + 1)] - ref0)
            assert np.mean(resultDiff) < 0.02
            assert np.max(resultDiff) < 1
            pass

    test_rfftStftConfig_01()


def test_MultichannelMultiResStftLayer():
    def _get_ref_output_single_res(time_sig, fft_size, frame_size, frame_shift, window_name, frame_nr, channel_nr):
        import numpy as np

        frame_start = frame_nr * frame_shift
        frame_end = frame_start + frame_size
        frame = time_sig[0, frame_start:frame_end, channel_nr]
        if window_name == "hanning":
            window = np.hanning(frame_size)
        windowed_frame = window * frame
        out = np.fft.rfft(windowed_frame, fft_size)
        return out

    def test_stftConfig_single_res_01():
        with make_scope() as session:
            layer_name = "stft_layer"
            fft_sizes = [400]
            frame_sizes = [400]
            frame_shift = 160
            window = "hanning"
            test_input = np.ones((1, 32000, 2), dtype=np.float32)
            num_outputs = (int(fft_sizes[0] / 2) + 1) * test_input.shape[2]
            config = Config()
            config.update(
                {
                    "num_outputs": num_outputs,
                    "num_inputs": test_input.shape[2],
                    "network": {
                        layer_name: {
                            "class": "multichannel_multiresolution_stft_layer",
                            "frame_shift": frame_shift,
                            "frame_sizes": frame_sizes,
                            "window": window,
                            "fft_sizes": fft_sizes,
                            "use_rfft": True,
                            "nr_of_channels": 2,
                            "is_output_layer": True,
                            "from": "data:data",
                        }
                    },
                }
            )
            network = TFNetwork(config=config, train_flag=True)
            network.construct_from_dict(config.typed_value("network"))
            layer = network.layers[layer_name]
            test_output = session.run(
                layer.output.placeholder, {network.get_extern_data("data").placeholder: test_input}
            )
            ref0 = _get_ref_output_single_res(test_input, fft_sizes[0], frame_sizes[0], frame_shift, window, 0, 0)
            resultDiff = np.abs(test_output[0, 0, 0 : (int(fft_sizes[0] / 2) + 1)] - ref0)
            assert test_output.shape[2] == num_outputs
            assert np.mean(resultDiff) < 0.02
            assert np.max(resultDiff) < 1

    def test_stftConfig_multi_res_01():
        with make_scope() as session:
            layer_name = "stft_layer"
            fft_sizes = [400, 200]
            frame_sizes = [400, 200]
            frame_shift = 160
            window = "hanning"
            test_input = np.ones((1, 32000, 2), dtype=np.float32)
            test_input[0, 1000:1100, 1] = np.ones((100), dtype=np.float32) * 0.5
            num_outputs = int(np.sum([(int(fft_size / 2) + 1) * test_input.shape[2] for fft_size in fft_sizes]))
            config = Config()
            config.update(
                {
                    "num_outputs": num_outputs,
                    "num_inputs": test_input.shape[2],
                    "network": {
                        layer_name: {
                            "class": "multichannel_multiresolution_stft_layer",
                            "frame_shift": frame_shift,
                            "frame_sizes": frame_sizes,
                            "window": window,
                            "fft_sizes": fft_sizes,
                            "use_rfft": True,
                            "nr_of_channels": 2,
                            "is_output_layer": True,
                            "from": "data:data",
                        }
                    },
                }
            )
            network = TFNetwork(config=config, train_flag=True)
            network.construct_from_dict(config.typed_value("network"))
            layer = network.layers[layer_name]
            test_output = session.run(
                layer.output.placeholder, {network.get_extern_data("data").placeholder: test_input}
            )
            assert test_output.shape[2] == num_outputs
            comparison_frame = 6
            ref00 = _get_ref_output_single_res(
                test_input, fft_sizes[0], frame_sizes[0], frame_shift, window, comparison_frame, 0
            )
            ref01 = _get_ref_output_single_res(
                test_input, fft_sizes[0], frame_sizes[0], frame_shift, window, comparison_frame, 1
            )
            ref10 = _get_ref_output_single_res(
                test_input, fft_sizes[1], frame_sizes[1], frame_shift, window, comparison_frame, 0
            )
            ref11 = _get_ref_output_single_res(
                test_input, fft_sizes[1], frame_sizes[1], frame_shift, window, comparison_frame, 1
            )
            ref = np.concatenate([ref00, ref01, ref10, ref11], axis=0)
            resultDiff = np.abs(test_output[0, comparison_frame, :] - ref)
            assert np.mean(resultDiff) < 0.02
            assert np.max(resultDiff) < 1

    def test_stftConfig_multi_res_02():
        with make_scope() as session:
            layer_name = "stft_layer"
            fft_sizes = [400, 200, 800]
            frame_sizes = [400, 200, 800]
            frame_shift = 160
            window = "hanning"
            test_input = np.random.normal(0, 0.6, (1, 3200, 2))
            num_outputs = int(np.sum([(int(fft_size / 2) + 1) * test_input.shape[2] for fft_size in fft_sizes]))
            config = Config()
            config.update(
                {
                    "num_outputs": num_outputs,
                    "num_inputs": test_input.shape[2],
                    "network": {
                        layer_name: {
                            "class": "multichannel_multiresolution_stft_layer",
                            "frame_shift": frame_shift,
                            "frame_sizes": frame_sizes,
                            "window": window,
                            "fft_sizes": fft_sizes,
                            "use_rfft": True,
                            "nr_of_channels": 2,
                            "is_output_layer": True,
                            "from": "data:data",
                        }
                    },
                }
            )
            network = TFNetwork(config=config, train_flag=True)
            network.construct_from_dict(config.typed_value("network"))
            layer = network.layers[layer_name]
            test_output = session.run(
                layer.output.placeholder, {network.get_extern_data("data").placeholder: test_input}
            )
            assert test_output.shape[2] == num_outputs
            comparison_frame = 6
            ref00 = _get_ref_output_single_res(
                test_input, fft_sizes[0], frame_sizes[0], frame_shift, window, comparison_frame, 0
            )
            ref01 = _get_ref_output_single_res(
                test_input, fft_sizes[0], frame_sizes[0], frame_shift, window, comparison_frame, 1
            )
            ref10 = _get_ref_output_single_res(
                test_input, fft_sizes[1], frame_sizes[1], frame_shift, window, comparison_frame, 0
            )
            ref11 = _get_ref_output_single_res(
                test_input, fft_sizes[1], frame_sizes[1], frame_shift, window, comparison_frame, 1
            )
            ref20 = _get_ref_output_single_res(
                test_input, fft_sizes[2], frame_sizes[2], frame_shift, window, comparison_frame, 0
            )
            ref21 = _get_ref_output_single_res(
                test_input, fft_sizes[2], frame_sizes[2], frame_shift, window, comparison_frame, 1
            )
            ref = np.concatenate([ref00, ref01, ref10, ref11, ref20, ref21], axis=0)
            resultDiff = np.abs(test_output[0, comparison_frame, :] - ref)
            assert np.mean(resultDiff) < 0.06
            assert np.max(resultDiff) < 1

    test_stftConfig_single_res_01()
    test_stftConfig_multi_res_01()
    test_stftConfig_multi_res_02()


if __name__ == "__main__":
    better_exchook.install()
    if len(sys.argv) <= 1:
        for k, v in sorted(globals().items()):
            if k.startswith("test_"):
                print("-" * 40)
                print("Executing: %s" % k)
                try:
                    v()
                except unittest.SkipTest as exc:
                    print("SkipTest:", exc)
                print("-" * 40)
        print("Finished all tests.")
    else:
        assert len(sys.argv) >= 2
        for arg in sys.argv[1:]:
            print("Executing: %s" % arg)
            if arg in globals():
                globals()[arg]()  # assume function and execute
            else:
                eval(arg)  # assume Python code and execute



"""
Diagnostic functions for GPU information, failings, memory usage, etc.
"""

from __future__ import annotations
from typing import Optional, Union, List, TextIO
import os
import sys
import gc
import subprocess
import torch
from returnn.util.better_exchook import better_exchook
from returnn.util.basic import human_bytes_size


def print_available_devices(*, file: Optional[TextIO] = None):
    """
    Print available devices, GPU (CUDA or other), etc.

    :param file: where to print to. stdout by default
    """
    if file is None:
        file = sys.stdout
    cuda_visible_devs = None
    if "CUDA_VISIBLE_DEVICES" in os.environ:
        print("CUDA_VISIBLE_DEVICES is set to %r." % os.environ["CUDA_VISIBLE_DEVICES"], file=file)
        cuda_visible_devs = dict(enumerate([int(d) for d in os.environ["CUDA_VISIBLE_DEVICES"].split(",") if d]))
    else:
        if torch.cuda.is_available():
            print("CUDA_VISIBLE_DEVICES is not set.", file=file)

    if torch.cuda.is_available():
        print("Available CUDA devices:")
        count = torch.cuda.device_count()
        if cuda_visible_devs is not None and len(cuda_visible_devs) != count:
            print(
                f"(Mismatch between CUDA device count {count}"
                f" and CUDA_VISIBLE_DEVICES {cuda_visible_devs} count {len(cuda_visible_devs)}?)",
                file=file,
            )
        for i in range(count):
            print(f"  {i + 1}/{count}: cuda:{i}", file=file)
            props = torch.cuda.get_device_properties(i)
            print(f"       name: {props.name}", file=file)
            print(f"       total_memory: {human_bytes_size(props.total_memory)}", file=file)
            print(f"       capability: {props.major}.{props.minor}", file=file)
            if cuda_visible_devs is not None:
                if len(cuda_visible_devs) == count:
                    dev_idx_s = cuda_visible_devs[i]
                else:
                    dev_idx_s = "?"
            else:
                dev_idx_s = i
            print(f"       device_index: {dev_idx_s}", file=file)
        if not count:
            print("  (None)")
    else:
        print("(CUDA not available)")


def print_using_cuda_device_report(dev: Union[str, torch.device], *, file: Optional[TextIO] = None):
    """
    Theano and TensorFlow print sth like: Using gpu device 2: GeForce GTX 980 (...)
    Print in a similar format so that some scripts which grep our stdout work just as before.
    """
    if file is None:
        file = sys.stdout
    if isinstance(dev, str):
        dev = torch.device(dev)
    assert dev.type == "cuda", f"expected CUDA device, got {dev}"
    if dev.index is not None:
        idx = dev.index
    else:
        idx = torch.cuda.current_device()
    if "CUDA_VISIBLE_DEVICES" in os.environ:
        cuda_visible_devs = dict(enumerate([int(d) for d in os.environ["CUDA_VISIBLE_DEVICES"].split(",") if d]))
        idx_s = cuda_visible_devs.get(idx, torch.cuda.device_count() + idx)
    else:
        idx_s = idx
    print(f"Using gpu device {idx_s}:", torch.cuda.get_device_name(idx), file=file)
    free, total = torch.cuda.mem_get_info(idx)
    print(f"Total GPU {idx_s} memory {human_bytes_size(total)}, free {human_bytes_size(free)}", file=file)


def diagnose_no_gpu() -> List[str]:
    """
    Diagnose why we have no GPU.
    Print to stdout, but also prepare summary strings.

    :return: summary strings
    """
    # Currently we assume Nvidia CUDA here, but once we support other backends (e.g. ROCm),
    # first check which backend is most reasonable here.

    res = []
    print("CUDA_VISIBLE_DEVICES:", os.environ.get("CUDA_VISIBLE_DEVICES", None))
    print("LD_LIBRARY_PATH:", os.environ.get("LD_LIBRARY_PATH", None))

    try:
        torch.cuda.init()
    except Exception as exc:
        print("torch.cuda.init() failed:", exc)
        better_exchook(*sys.exc_info(), debugshell=False)
        res.append(f"torch.cuda.init() failed: {type(exc).__name__} {exc}")

    try:
        subprocess.check_call(["nvidia-smi"])
    except Exception as exc:
        print("nvidia-smi failed:", exc)
        better_exchook(*sys.exc_info(), debugshell=False)
        res.append(f"nvidia-smi failed")

    return res


def garbage_collect():
    """
    Perform garbage collection, including any special logic for GPU.

    Also see:
    https://github.com/pytorch/pytorch/issues/18853
    https://github.com/pytorch/pytorch/issues/27600
    https://pytorch.org/docs/stable/notes/faq.html#my-out-of-memory-exception-handler-can-t-allocate-memory
    https://github.com/Lightning-AI/pytorch-lightning/blob/7a4b0fc4331633cdf00b88776689e8a84ef96cb4/src/lightning/pytorch/utilities/memory.py#L83
    """
    gc.collect()
    if torch.cuda.is_initialized():
        torch.cuda.empty_cache()
        torch.cuda.reset_peak_memory_stats()
        stats = [
            f"alloc {human_bytes_size(torch.cuda.memory_allocated())}",
            f"reserved {human_bytes_size(torch.cuda.memory_reserved())}",
        ]
        print(f"CUDA memory usage after triggered GC:", " ".join(stats))



#!/usr/bin/env python3

"""
Given BPE codes/vocab, create a lexicon (mapping of words to all possible BPE seqs).
"""

from __future__ import annotations

from argparse import ArgumentParser
from xml.etree import ElementTree

import _setup_returnn_env  # noqa
from returnn.util import bpe as bpe_utils


def parse_vocab(filename):
    """
    Can be either pure text file, line-based, or lexicon XML file, or Python vocab dict.

    :param str filename:
    :rtype: list[str]
    """
    if filename.endswith(".gz"):
        import gzip

        raw = gzip.open(filename, "r").read().decode("utf8")
    else:
        raw = open(filename, "r").read()
    if raw.startswith("{"):  # Python dict (str sym -> int idx)
        py_vocab = eval(raw)
        assert isinstance(py_vocab, dict)  # sym -> idx
        labels = {idx: label for (label, idx) in sorted(py_vocab.items())}
        min_label, max_label, num_labels = min(labels), max(labels), len(labels)
        assert 0 == min_label
        if num_labels - 1 < max_label:
            print("Vocab error: not all indices used? max label: %i" % max_label)
            print("unused labels: %r" % ([i for i in range(max_label + 1) if i not in labels],))
        assert num_labels - 1 == max_label
        zero_sym = labels[0]
        assert isinstance(zero_sym, str)
        return [label for (idx, label) in sorted(labels.items())]
    if raw.startswith("<?xml"):  # lexicon XML
        labels = []
        from io import StringIO

        raw_stream = StringIO(raw)
        context = iter(ElementTree.iterparse(raw_stream, events=("start", "end")))
        _, root = next(context)  # get root element
        for event, elem in context:
            if event == "end" and elem.tag == "lemma":
                for orth_elem in elem.findall("orth"):
                    orth = (orth_elem.text or "").strip()
                    labels.append(orth)
                root.clear()  # free memory
        return labels
    # Assume line-based. No idea how to to a good sanity check...
    return raw.splitlines()


def xml_prettify(element, indent="  "):
    """
    https://stackoverflow.com/a/38574067/133374 (deleted StackOverflow answer)

    :param ElementTree.Element element:
    :param str indent:
    """
    queue = [(0, element)]  # (level, element)
    while queue:
        level, element = queue.pop(0)
        children = [(level + 1, child) for child in list(element)]
        if children:
            element.text = "\n" + indent * (level + 1)  # for child open
        if queue:
            element.tail = "\n" + indent * queue[0][0]  # for sibling open
        else:
            element.tail = "\n" + indent * (level - 1)  # for parent close
        queue[0:0] = children  # prepend so children come before siblings


def main():
    """
    Main entry.
    """
    arg_parser = ArgumentParser()
    arg_parser.add_argument("--bpe_vocab", required=True)
    arg_parser.add_argument("--word_vocab", required=True)
    arg_parser.add_argument("--unk")
    arg_parser.add_argument("--skip_special", action="store_true")
    arg_parser.add_argument("--lower_case", action="store_true")
    arg_parser.add_argument("--output")
    args = arg_parser.parse_args()

    bpe_syms = parse_vocab(args.bpe_vocab)
    words = parse_vocab(args.word_vocab)
    print("BPE symbols: num %i, first %r" % (len(bpe_syms), bpe_syms[0]))
    print("Words: num %i, first %r" % (len(words), words[0]))

    print("Build BPE prefix tree...")
    bpe = bpe_utils.PrefixTree()
    for bpe_sym in bpe_syms:
        bpe.add(bpe_sym)

    print("Build lexicon...")
    xml = ElementTree.Element("lexicon")
    xml_phone_inventory = ElementTree.SubElement(xml, "phoneme-inventory")
    for bpe_sym in bpe_syms:  # each BPE symbol will be a phoneme in the XML
        xml_phone = ElementTree.SubElement(xml_phone_inventory, "phoneme")
        ElementTree.SubElement(xml_phone, "symbol").text = bpe_sym
        ElementTree.SubElement(xml_phone, "variation").text = "context"

    visited_words = set()

    # noinspection PyShadowingNames
    def visit_word(word):
        """
        :param str word:
        """
        if word in visited_words:
            return
        visited_words.add(word)
        bpe_sym_seqs = bpe_utils.CharSyncSearch(bpe=bpe, word=word).search()
        if not bpe_sym_seqs:
            print("no BPE seq found for word %r" % word)
            return
        xml_lemma = ElementTree.SubElement(xml, "lemma")
        ElementTree.SubElement(xml_lemma, "orth").text = word
        for bpe_sym_seq in bpe_sym_seqs:
            ElementTree.SubElement(xml_lemma, "phon").text = " ".join(bpe_sym_seq)

    for word in words:
        if args.lower_case:
            word = word.lower()
        if not word:
            continue
        if args.skip_special:
            if word.startswith("[") and word.endswith("]"):
                continue
            if word.startswith("<") and word.endswith(">"):
                continue
        visit_word(word)

    for bpe_sym in bpe_syms:
        if bpe_sym.endswith(bpe_utils.BpeMergeSymbol):
            continue
        if bpe_sym not in words:
            continue
        # E.g. special symbols, which were skipped above.
        visit_word(bpe_sym)

    if args.output:
        xml_prettify(xml)
        xml_str = ElementTree.tostring(xml, encoding="utf-8")
        with open(args.output, "wb") as f:
            f.write(xml_str)
        print("Wrote XML:", args.output)
    else:
        print("Specify --output to save the XML.")


if __name__ == "__main__":
    from returnn.util import better_exchook

    better_exchook.install()
    main()



"""
Public backend-related functions.
"""

from __future__ import annotations

from . import _backend

# And some functions from the internal backend API.
from ._backend import select_backend_torch, select_backend_returnn_layers_tf

__all__ = ["select_backend_torch", "select_backend_returnn_layers_tf", "is_backend_raw_tensor_dim_tag_independent"]


def is_backend_raw_tensor_dim_tag_independent() -> bool:
    """
    :return: whether raw tensors of the backend are independent of :class:`Dim`
        (Usually yes, e.g. :class:`tf.Tensor` or :class:`torch.Tensor`,
        but the TF-layers backend is an exception.)
    """
    return _backend.global_backend.is_backend_raw_tensor_dim_tag_independent



"""
TensorFlow frontend.
"""

from ._backend import ReturnnLayersBackend

# Sorted by name.
from .cond import *
from .config_entry_points import *
from .debug_eager_mode import *
from .dims import *
from .layer import *
from .loop import *
from .make_layer import *
from .masked_computation import *
from .prev_tensor_ref import *


# https://returnn.readthedocs.io/en/latest/configuration_reference/behavior_version.html
# Need Dim.is_equal to be more restrictive (v16).
min_returnn_behavior_version = 16
